@@ -8,32 +8,59 @@ use POSIX;
 use Config::Tiny;
 use POE qw( Wheel::Run );
 use Term::ANSIColor qw(:constants);
+use Text::Balanced;
 
-our $VERSION = '2.0000';
-our @DEFAULT_SHELL_COMMAND = (ssh => '-o', 'BatchMode yes', '-o', 'StrictHostKeyChecking no', '-o', 'ConnectTimeout 20', '%h');
+our $VERSION = '2.0207_1';
+our @DEFAULT_SHELL_COMMAND = (ssh => '-o', 'BatchMode yes', '-o', 'StrictHostKeyChecking no', '-o', 'ConnectTimeout 20', '[%u]-l', '[]%u', '%h');
 
 # new {{{
 sub new {
-    my $this = bless { hosts=>[], cmd=>[], _shell_cmd=>[@DEFAULT_SHELL_COMMAND] };
+    my $class = shift;
+    my $this  = bless { hosts=>[], cmd=>[], _shell_cmd=>[@DEFAULT_SHELL_COMMAND] }, $class;
 
-    $this;
+    return $this;
 }
 # }}}
 
 # _process_space_delimited {{{
 sub _process_space_delimited {
     my $this = shift;
+    my $that = shift;
 
-    return
-        grep {defined} ($_[0] =~ m/["']([^"']*?)["']|(\S+)/g)
+    my @output;
+    while( $that ) {
+        if( $that =~ m/^\s*['"]/ ) {
+            my ($tok, $rem) = Text::Balanced::extract_delimited($that, qr(["']));
+
+            ($tok =~ s/^(['"])// and $tok =~ s/$1$//) or die "internal error processing space delimited";
+
+            push @output, $tok;
+            $that = $rem;
+
+        } else {
+            my ($tok, $rem) = split ' ', $that, 2; 
+
+            push @output, $tok;
+            $that = $rem;
+        }
+    }
+
+    return @output
 }
 # }}}
 # _process_hosts {{{
 sub _process_hosts {
     my $this = shift;
+
     my @h = do {
         my @tmp = map { my $k = $_; $k =~ s/^\@// ? @{$this->{groups}{$k} or die "couldn't find group: \@$k\n"} : $_ } @_;
         my %h; @h{@tmp} = ();
+        for(keys %h) {
+            if(my ($k) = m/^\-(.+)/) {
+                delete $h{$_};
+                delete $h{$k};
+            }
+        }
         sort keys %h;
     };
 
@@ -41,16 +68,17 @@ sub _process_hosts {
     for( map { length $this->_host_route_to_nick($_) } @h ) {
         $l = $_ if $_>$l
     }
+
     $this->{_host_width} = $l if $l != $o;
 
-    @h;
+    return @h;
 }
 # }}}
 # _host_route_to_nick {{{
 sub _host_route_to_nick {
     my $this = shift;
 
-    join "", shift =~ m/(?:!|[^!]+$)/g
+    return join "", shift =~ m/(?:!|[^!]+$)/g
 }
 # }}}
 
@@ -66,24 +94,57 @@ sub set_shell_command_option {
         $this->{_shell_cmd} = [ $this->_process_space_delimited($arg||"") ];
     }
 
-    $this;
+    return $this;
 }
 # }}}
 # set_group_option {{{
 sub set_group_option {
-    my $this  = shift;
+    my $this   = shift;
+    my $groups = ($this->{groups} ||= {});
 
     my ($name, $value);
     while( ($name, $value) = splice @_, 0, 2 and $name and $value ) {
         if( ref($value) eq "ARRAY" ) {
-            $this->{groups}{$name} = [ @$value ]; # make a real copy
+            $groups->{$name} = [ @$value ]; # make a real copy
 
         } else {
-            $this->{groups}{$name} = [ $this->_process_space_delimited( $value ) ];
+            $groups->{$name} = [ $this->_process_space_delimited( $value ) ];
         }
     }
 
-    $this;
+    my @groups = keys %{ $this->{groups} };
+    my $replace_limit = 30;
+    REPLACE_GROPUS: {
+        my $replaced = 0;
+
+        for my $group (@groups) {
+            my $hosts = $groups->{$group};
+
+            my $r = 0;
+            for(@$hosts) {
+                if( m/^@(.+)/ ) {
+                    if( my $g = $groups->{$1} ) {
+                        $_ = $g;
+
+                        $r ++;
+                    }
+                }
+            }
+
+            if( $r ) {
+                my %h;
+                @h{ map {ref $_ ? @$_ : $_} @$hosts } = ();
+                $groups->{$group} = [ keys %h ];
+                $replaced ++;
+            }
+        }
+
+        $replace_limit --;
+        last if $replace_limit < 1;
+        redo if $replaced;
+    }
+
+    return $this;
 }
 # }}}
 # set_logfile_option {{{
@@ -93,7 +154,7 @@ sub set_logfile_option {
     my $trunc = shift;
 
     unless( our $already_compiled++ ) {
-        eval q {
+        my $load_ansi_filter_package = q {
             package App::MrShell::ANSIFilter;
             use Symbol;
             use Tie::Handle;
@@ -115,13 +176,17 @@ sub set_logfile_option {
                 $pfft;
             }
 
-        1} or die $@;
+        1};
+
+        eval $load_ansi_filter_package or die $@; ## no critic -- sometimes this kind of eval is ok
+        # (This probably isn't one of them.)
     }
 
-    open my $log, ($trunc ? ">" : ">>"), $file or croak "couldn't open $file for write: $!";
+    open my $log, ($trunc ? ">" : ">>"), $file or croak "couldn't open $file for write: $!"; ## no critic -- I mean to pass this around, shut up
 
     $this->{_log_fh} = App::MrShell::ANSIFilter::filtered_handle($log);
-    $this;
+
+    return $this;
 }
 # }}}
 # set_debug_option {{{
@@ -139,28 +204,43 @@ sub set_debug_option {
     }
 
     $this->{debug} = $val ? $val : 1;
-    $this;
+
+    return $this;
 }
 # }}}
 # set_no_command_escapes_option {{{
 sub set_no_command_escapes_option {
     my $this = shift;
 
-    $this->{no_command_escapes} = 1;
-    $this;
+    $this->{no_command_escapes} = shift || 0;
+
+    return $this;
+}
+# }}}
+
+# groups {{{
+sub groups {
+    my $this = shift;
+
+    return unless $this->{groups};
+    return wantarray ? %{$this->{groups}} : $this->{groups};
 }
 # }}}
 
 # set_usage_error($&) {{{
-sub set_usage_error($&) {
+sub set_usage_error($&) { ## no critic -- prototypes are bad how again?
     my $this = shift;
     my $func = shift;
     my $pack = caller;
     my $name = $pack . "::$func";
     my @args = @_;
 
-    $this->{_usage_error} = sub { no strict 'refs'; $name->(@args) };
-    $this;
+    $this->{_usage_error} = sub {
+        no strict 'refs'; ## no critic -- how would you call this by name without this?
+        $name->(@args)
+    };
+
+    return $this;
 }
 # }}}
 # read_config {{{
@@ -174,25 +254,25 @@ sub read_config {
     }
 
     if( my $c = $this->{_conf}{options}{'shell-command'} ) {
-        $this->set_shell_command_option( 1, $c );
+        $this->set_shell_command_option( $c );
     }
 
     if( my $c = $this->{_conf}{options}{'logfile'} ) {
         my $t = $this->{_conf}{options}{'truncate-logfile'};
-        my $v = 1 if $t;
-           $v = 0 if $c =~ m/(?:no|false)/i;
+        my $v = ($t ? 1:0);
+           $v = 0 if $t =~ m/(?:no|false)/i;
 
-        $this->set_logfile_option($c, $t);
+        $this->set_logfile_option($c, $v);
     }
 
     if( my $c = $this->{_conf}{options}{'no-command-escapes'} ) {
-        my $v = 1 if $c;
+        my $v = ($c ? 1:0);
            $v = 0 if $c =~ m/(?:no|false)/i;
 
         $this->set_no_command_escapes_option( $v );
     }
 
-    $this;
+    return $this;
 }
 # }}}
 # set_hosts {{{
@@ -200,7 +280,8 @@ sub set_hosts {
     my $this = shift;
 
     $this->{hosts} = [ $this->_process_hosts(@_) ];
-    $this;
+
+    return $this;
 }
 # }}}
 # queue_command {{{
@@ -227,7 +308,7 @@ sub queue_command {
         push @{$this->{_cmd_queue}{$h}}, [@_]; # make a real copy
     }
 
-    $this;
+    return $this;
 }
 # }}}
 # run_queue {{{
@@ -245,7 +326,7 @@ sub run_queue {
 
     POE::Kernel->run();
 
-    $this
+    return $this;
 }
 # }}}
 
@@ -261,7 +342,7 @@ sub std_msg {
     my $time_str = strftime('%H:%M:%S', localtime);
 
     print $time_str,
-        sprintf(' %4s %-*s', "[$cmdno]", $this->{_host_width}+2, $host_msg),
+        sprintf(' %-*s', $this->{_host_width}+2, $host_msg),
             ( $fh==2 ? ('[',BOLD,YELLOW,'stderr',RESET,'] ') : () ), $msg, RESET, "\n";
 
     if( $this->{_log_fh} ) {
@@ -270,10 +351,10 @@ sub std_msg {
         # No point in printing colors, stripped anyway.  Formatting columns is
         # equally silly -- in append mode anyway.
         $host_msg = $host ? "$host: " : "";
-        print {$this->{_log_fh}} "$time_str [$cmdno] $host_msg", ($fh==2 ? "[stderr] " : ""), $msg, "\n";
+        print {$this->{_log_fh}} "$time_str $host_msg", ($fh==2 ? "[stderr] " : ""), $msg, "\n";
     }
 
-    $this;
+    return $this;
 }
 # }}}
 
@@ -286,6 +367,8 @@ sub line {
 
     $$lineno ++;
     $this->std_msg($host, $cmdno, $fh, $line);
+
+    return;
 }
 # }}}
 
@@ -296,6 +379,8 @@ sub _sigchld_exit_error {
     $exit >>= 8;
 
     $this->std_msg("?", -1, 0, BOLD.RED."-- sigchld received for untracked pid($pid, $exit), probably a bug in Mr. Shell --");
+
+    return;
 }
 
 sub sigchld {
@@ -321,6 +406,8 @@ sub sigchld {
     }
 
     $_[KERNEL]->yield( stall_close => $kid->ID, 0 );
+
+    return;
 }
 # }}}
 # _close {{{
@@ -349,6 +436,8 @@ sub _close {
     } else {
         $_[KERNEL]->yield( stall_close => $wid, $count+1 );
     }
+
+    return;
 }
 # }}}
 # error_event {{{
@@ -358,8 +447,10 @@ sub error_event {
     my ($kid, $host, $cmdno, @c) = @{ delete $this->{_wid}{$wid} || return };
     delete $this->{_pid}{ $kid->PID };
 
-    $errstr = "remote end closed" if $operation eq "read" and !$errnum;
+    $errstr = "remote end closed" if $operation eq "read" and not $errnum;
     $this->std_msg($host, $cmdno, 0, RED."-- $operation error $errnum: $errstr --");
+
+    return;
 }
 # }}}
 
@@ -371,7 +462,7 @@ sub set_subst_vars {
         $this->{_subst}{$k} = $v unless exists $this->{_subst}{$k};
     }
 
-    $this;
+    return $this;
 }
 # }}}
 # subst_cmd_vars {{{
@@ -385,15 +476,18 @@ sub subst_cmd_vars {
                 #  proovably true that I do.)
 
     if( $host =~ m/\b(?!<\\)!/ ) {
-        delete $h{'%h'};
         my @hosts = split '!', $host;
 
+        my @indexes_of_replacements;
         for(my $i=0; $i<@c; $i++) {
             if( $c[$i] eq '%h' ) {
                 splice @c, $i, 1, $hosts[0];
 
+                push @indexes_of_replacements, $i;
+
                 for my $h (reverse @hosts[1 .. $#hosts]) {
                     splice @c, $i+1, 0, @c[0 .. $i-1] => $h;
+                    push @indexes_of_replacements, $i+1 + $indexes_of_replacements[-1];
 
                     unless( $this->{no_command_escapes} ) {
                         for my $arg (@c[$i+1 .. $#c]) {
@@ -412,20 +506,45 @@ sub subst_cmd_vars {
             }
         }
 
+        my $beg = 0;
+        for my $i (@indexes_of_replacements) {
+            if( $c[$i] =~ s/^([\w.\-_]+)@// ) {
+                my $u = $1;
+                for(@c[$beg .. $i-1]) {
+                    s/^(\[\%u\]|\[\](?=\%u))//;
+                    $_ = $u if $_ eq '%u';
+                }
+
+            } else {
+                # NOTE: there's really no need to go through and remove [%u]
+                # conditional options, they'll automatically get nuked below
+                $c[$i] =~ s/\\@/@/g;
+            }
+            $beg = $i+1;
+        }
+
+        delete $h{'%h'};
+
     } else {
         $h{'%h'} =~ s/\\!/!/g;
     }
 
-    if( $this->{debug} ) {
-        my @cmd = map {exists $h{$_} ? $h{$_} : $_} @c;
-
-        my @dt = map {"<$_>"} @cmd;
-        $this->std_msg($host, $h{'%n'}, 0, BOLD.BLACK."DEBUG: exec(@dt)");
-
-        return @cmd;
+    if( $h{'%h'} ) {
+        $h{'%u'} = $1 if $h{'%h'} =~ s/^([\w.\-_]+)@//;
+        $h{'%h'} =~ s/\\@/@/g;
     }
 
-    map {exists $h{$_} ? $h{$_} : $_} @c;
+    @c = map {exists $h{$_} ? $h{$_} : $_}
+         map { m/^\[([^\[\]]+)\]/ ? ($h{$1} ? do{s/^\[\Q$1\E\]//; $_} : ()) : ($_) } ## no critic: why on earth not?
+         map { s/\[\]\%(\w+)/[\%$1]\%$1/; $_ }                                       ## no critic: why on earth not?
+         @c;
+
+    if( $this->{debug} ) {
+        local $" = ")(";
+        $this->std_msg($host, $h{'%n'}, 0, BOLD.BLACK."DEBUG: exec(@c)");
+    }
+
+    return @c;
 }
 # }}}
 # start_queue_on_host {{{
@@ -439,7 +558,7 @@ sub start_queue_on_host {
     );
 
     my $kid = POE::Wheel::Run->new(
-        Program     => [ $this->subst_cmd_vars(@{$this->{_shell_cmd}} => @$cmd) ],
+        Program     => [ my @debug_rq = ($this->subst_cmd_vars(@{$this->{_shell_cmd}} => @$cmd)) ],
         StdoutEvent => "child_stdout",
         StderrEvent => "child_stderr",
         CloseEvent  => "child_close",
@@ -450,6 +569,8 @@ sub start_queue_on_host {
     my $lineno = 0;
     my $info = [ $kid, $host, $cmdno, \$lineno, @next ];
     $this->{_wid}{ $kid->ID } = $this->{_pid}{ $kid->PID } = $info;
+
+    return;
 }
 # }}}
 
@@ -478,6 +599,9 @@ sub poe_start {
     }
 
     delete $this->{_cmd_queue};
+
     return;
 }
 # }}}
+
+1;
